/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import javax.lang.model.type.NullType;
import javax.swing.ImageIcon;
import javax.swing.SwingWorker;
import nl.cloudfarming.client.field.model.ActivityField;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DataLog;
import nl.cloudfarming.client.isobus.model.GridTypeEnum;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import nl.cloudfarming.client.isobus.model.IsobusManager;
import nl.cloudfarming.client.isobus.model.IsobusMarshaller;
import nl.cloudfarming.client.isobus.model.IsobusTask;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTaskStatusEnum;
import nl.cloudfarming.client.isobus.model.IsobusTreatmentZone;
import nl.cloudfarming.client.isobus.model.ProcessDataVariable;
import nl.cloudfarming.client.isobus.model.ValuePresentation;
import nl.cloudfarming.client.isobus.model.util.IsobusObjectFactoryUtil;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.task.ExportableTask;
import nl.cloudfarming.client.task.grid.TreatmentZoneGrid;
import nl.cloudfarming.client.util.Callback;
import org.apache.commons.lang.StringUtils;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.netbeans.api.project.Project;
import org.openide.awt.NotificationDisplayer;
import org.openide.filesystems.FileObject;
import org.openide.util.Exceptions;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;

/**
 * Implementation of isobusManager Is able to import and export TASKDATA.XML
 * files
 *
 * @author Merijn Zengers
 */
@ServiceProvider(service = IsobusManager.class)
@NbBundle.Messages({
    "IsobusManagerImpl treatment zone limit exceeded=Maximum number of treatment zones exceeded, exporting first 254 treatment zones!",
    "IsobusManagerImpl generating isobus export=Exporting isobus",
    "IsobusManagerImpl generating export done title=Exporting complete",
    "IsobusManagerImpl generating export done full description=Exporting of the isobus file is completed"
})
public class IsobusManagerImpl implements IsobusManager {

    public final static String ISOBUS_FOLDER = "ISO11783";
    private static final AppLogger LOG = AppLogFactory.getLogger(IsobusManagerImpl.class);
    private static final String GRID_FILE_PREFIX = "GRD";
    private static final int GRID_ID_LENGHT = 5;
    private static final short EMPTY_TREATMENTZONE_CODE = 0;
    private static final String EMPTY_TREATMENTZONE_DESIGNATOR = "Empty";
    private static final short EMPTY_TREATMENTZONE_COLOR = 0;
    private static final short DEFAULT_TREATMENTZONE_CODE = 1;
    private static final String DEFAULT_TREATMENTZONE_DESIGNATOR = "Default";
    private static final short DEFAULT_TREATMENTZONE_COLOR = 0;
    private static final String DATA_TRANSFER_ORIGIN = "1";
    private static final String MANAGEMENT_SOFTWARE = "AgroSense";
    private static final String VALUE_PRESENTATION_ID = "VPN1";
    //TODO: get VPN details from task or activity field
    private static final int VALUE_PRESENTATION_OFFSET = 0;
    private static final float VALUE_PRESENTATION_SCALE = 1.0f;
    private static final short VALUE_PRESENTATION_DECIMALS = 0;
    private static final String VALUE_PRESENTATION_DESIGNATOR = "aan/uit";
    private static final String ISOBUS_TASK_ID_PREFIX = "TSK";
    private static final IsobusTaskStatusEnum TASK_STATUS = IsobusTaskStatusEnum.INITIAL;
    private static final String GRID_FILE_EXT = "BIN";
    private static final short TREATMENTZONE_TYPE_2_CODE = 2;
    private static final String TREATMENT_ZONE_TYPE_2_NAME = "Precision application";
    private ImageIcon INFO_ICON = ImageUtilities.loadImageIcon("nl/cloudfarming/client/isobus/type/information.png", true);

    /**
     * Get all the taskdata folders from the provided project These folders
     * contain a TASKDATA.XML and optional GRD.BIN ,TLG.BIN and TLG.XML files
     *
     * @param project the project to search in
     * @return the found folders or empty array if there are no children
     */
    @Override
    public FileObject[] getTaskDataFolders(Project project) {
        FileObject isobusFolder = getIsobusFolder(project);
        return isobusFolder.getChildren();
    }

    /**
     * Returns isobusfolder inside the specified project Protected for
     * unittesting purposes!
     *
     * @param project Project to get isobusfolder from.
     * @return Isobusfolder on the filesystem
     */
    @Override
    public FileObject getIsobusFolder(Project project) {
        FileObject isobusFolder = project.getProjectDirectory().getFileObject(ISOBUS_FOLDER);
        // Create Isobusfolder if it doesn't exist
        if (isobusFolder == null) {
            try {
                isobusFolder = project.getProjectDirectory().createFolder(ISOBUS_FOLDER);
            } catch (IOException ex) {
                Exceptions.printStackTrace(ex);
            }
        }
        return isobusFolder;
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public void exportTaskdata(final FileObject destFolder, final GridTypeEnum gridType, final DDI ddi, final double multiplier, final Callback<Object> callback, final ExportableTask... tasks) {
        //This long block of code contains several long operations. Placing each specific
        //method call in a separate thread would be very tricky and would result in
        //a very fragile threading construction. That's why this entire block
        //has been place in one SwingWorker that makes sure the UI thread is not
        //disturbed.
        SwingWorker worker = new SwingWorker() {

            @Override
            protected NullType doInBackground() throws Exception {
                ProgressHandle progress = ProgressHandleFactory.createHandle(Bundle.IsobusManagerImpl_generating_isobus_export());
                progress.start();
                IsobusTaskData isobusTaskData = createIsobusTaskData();
                ValuePresentation presentation = createValuePresentation();
                isobusTaskData.getVPN().add(presentation);

                int gridCounter = 0;
                // loop through all tasks and create isobus tasks
                for (ExportableTask task : tasks) {
                    ActivityField activityField = task.getLookup().lookup(ActivityField.class);
                    IsobusTask isobusTask = createIsobusTask(task.getURI().toString());
                    //Task IDs are only valid if they start with 'Task' and are
                    //followed by either either a '-' and one or several numbers
                    //or 'Task' followed by one or several numbers. Like so:
                    //TSK12345
                    //TSK2
                    //TSK-998
                    //TSK-2
                    isobusTask.setTaskId(ISOBUS_TASK_ID_PREFIX + gridCounter);

                    DataLog dataLog = createDataLog();
                    isobusTask.getDLT().add(dataLog);

                    IsobusTreatmentZone emptyTreatmentZone = createEmptyTreatmentZone();
                    emptyTreatmentZone.getPDV().add(createProcessDataVariable(presentation, ddi));
                    isobusTask.getTZN().add(emptyTreatmentZone);
                    // TODO: why not set empty TZN as out of field TZN?
                    // isobusTask.setOutOfFieldTreatmentZoneCode(emptyTreatmentZone.getTreatmentZoneCode());
                    
                    IsobusTreatmentZone defaultTreatmentZone = createDefaultTreatmentZone();
                    // TODO: set PDV value to activityField average (possibly weighted using TZN areas);
                    // that wouldn't make sense for the on/off DDI though
                    defaultTreatmentZone.getPDV().add(createProcessDataVariable(presentation, ddi));
                    isobusTask.getTZN().add(defaultTreatmentZone);
                    isobusTask.setDefaultTreatmentZoneCode(defaultTreatmentZone.getTreatmentZoneCode());
                    // FIXME: addition of a default TZN on top of the empty TZN implies changes
                    // needed to max-number-of-tzn constants in other modules (254->253) 

                    isobusTask.setTaskDesignator(task.getName());

                    // TSK
                    isobusTaskData.getTSK().add(isobusTask);

                    TreatmentZoneGrid grid = task.getExportableGrid();

                    isobusTask.getTZN().addAll(createTreatmentZones(activityField.getTreatmentZones(), presentation, grid, gridType, ddi, multiplier));

                    // Gridfilename should be placed in taskdata without extension
                    String gridFileName = buildGridFileName(gridCounter);

                    IsobusGrid isobusGrid = createIsobusGrid(isobusTask.getTZN(), grid, gridType, gridFileName);
                    isobusTask.setGRD(isobusGrid);

                    //Start write the grid file
                    GridWriter gridWriter = new GridWriter();
                    // remove old file first if it exists:
                    FileObject gridFile = destFolder.getFileObject(isobusGrid.getFilename(), GRID_FILE_EXT);
                    if (gridFile != null) {
                        gridFile.delete();
                    }
                    gridFile = destFolder.createData(isobusGrid.getFilename(), GRID_FILE_EXT);
                    gridWriter.write(gridFile, gridType, grid, ddi, multiplier);
                    //End write the grid file

                    //raise the grid counter so the next task
                    //will have its own grid file
                    gridCounter++;
                }
                //FIXME needs some heavy refactoring! why pass on grid file separately? its in the task
                IsobusMarshaller isobusMarshaller = new IsobusMarshaller();
                isobusMarshaller.marshall(destFolder, isobusTaskData);
                progress.finish();
                if (callback != null) {
                    callback.call(null);
                }
                NotificationDisplayer.getDefault().notify(Bundle.IsobusManagerImpl_generating_export_done_title(), INFO_ICON, Bundle.IsobusManagerImpl_generating_export_done_full_description(), null);
                return null;
            }
        };
        worker.execute();
    }

    /**
     * <p>Create a {@link IsobusGrid}</p> <p>Can create grids for {@link GridTypeEnum#TYPE_1}
     * and {@link GridTypeEnum#TYPE_2}. For type one will set no {@link IsobusGrid#treatmentZoneCode}.
     * For type 2 will set the code of the prototype {@link IsobusTreatmentZone} in
     * the provided list</p>
     *
     * @param treatmentZones if list is empty and {@link GridTypeEnum#TYPE_2} is passed will throw an assertion exception 
     * @param grid
     * @param gridType
     * @param gridFileName
     * @return the Created isobusGrid
     */
    IsobusGrid createIsobusGrid(List<IsobusTreatmentZone> treatmentZones, TreatmentZoneGrid grid, GridTypeEnum gridType, String gridFileName) {
        IsobusGrid isobusGrid = null;
        if (GridTypeEnum.TYPE_1.equals(gridType)) {
            isobusGrid = IsobusObjectFactoryUtil.createGrid(grid, gridType, gridFileName, null);
        } else if (GridTypeEnum.TYPE_2.equals(gridType)) {
            assert !treatmentZones.isEmpty() : "Cannot create a type 2 IsobusGrid without treatmentzones";
            // TZN 0: the empty tzn, TZN 1: the prototype tzn
            isobusGrid = IsobusObjectFactoryUtil.createGrid(grid, gridType, gridFileName, treatmentZones.get(2).getTreatmentZoneCode());
        }
        return isobusGrid;
    }

    /**
     * <p>Create one or more {@link IsobusTreatmentZone} from the provided list
     * of {@link TreatmentZone}</p> <p>Can create list for both type of {@link GridTypeEnum}.
     * For {@link GridTypeEnum#TYPE_1} creates a {@link IsobusTreatmentZone}
     * with one {@link ProcessDataVariable} for every {@link TreatmentZone} in
     * the passed list. For {@link GridTypeEnum#TYPE_2} will create one {@link IsobusTreatmentZone}
     * with a {@link ProcessDataVariable} for every {@link TreatmentZone} in the
     * passed list</p>
     *
     * @param treatmentZones
     * @param presentation
     * @param grid
     * @param gridType
     * @return
     */
    List<IsobusTreatmentZone> createTreatmentZones(List<TreatmentZone> treatmentZones, ValuePresentation presentation, TreatmentZoneGrid grid, GridTypeEnum gridType, DDI ddi, double multiplier) {
        List<IsobusTreatmentZone> created = new ArrayList<>();

        switch(gridType) {
            case TYPE_1:
                // loop through treatmentzones in an agrosense task, convert them to 
                // isobus treatmentzones and add them to the task object (via TZN node)
                int counter = 2;
                for (TreatmentZone treatmentZone : treatmentZones) {
                    IsobusTreatmentZone isobusTreatmentZone = IsobusObjectFactoryUtil.createIsobusTreatmentZone(presentation, treatmentZone, grid.getTreatmentZoneCode(treatmentZone), ddi, multiplier);
                    if (isobusTreatmentZone != null) {
                        created.add(isobusTreatmentZone);
                    }
                    //FIXME: ISOBUS supports a maximum number of 255  treatment zones 
                    // (including the initial empty one, and the default one). In order to support more 
                    // treatment zones they should probably be merged beforehand.
                    if (counter == 254) {
                        LOG.severe(Bundle.IsobusManagerImpl_treatment_zone_limit_exceeded());
                        break;
                    }
                    counter++;
                }
                break;
            case TYPE_2:
                // create a single prototype TZN:
                IsobusTreatmentZone isobusTreatmentZone = new IsobusTreatmentZone();
                isobusTreatmentZone.setTreatmentZoneDesignator(TREATMENT_ZONE_TYPE_2_NAME);
                isobusTreatmentZone.setTreatmentZoneCode(TREATMENTZONE_TYPE_2_CODE);
                // add all used variables to the prototype without a value (i.e. set to 0, it's a required attribute)
                // Agrosense currently only supports a single variable:
                isobusTreatmentZone.getPDV().add(IsobusObjectFactoryUtil.createProcessDataVariable(0, presentation, ddi, multiplier));
                created.add(isobusTreatmentZone);
                break;
            default:
                throw new IllegalArgumentException("Unsupported isobus grid type: " + gridType);
        }
        return created;
    }

    /**
     * Generate a Gridfilename using the specified integer. For example:
     * GRD00005 Gridfilename stored in taskdata.xml should not have an extension
     * according to the ISOBUS standard.
     *
     * Protected for testing pusposes only.
     *
     * @param counter
     * @return Generated gridfilename
     */
    protected String buildGridFileName(int counter) {
        String counterValue = String.valueOf(counter++);
        StringBuilder sb = new StringBuilder(GRID_FILE_PREFIX);
        sb.append(StringUtils.leftPad(counterValue, GRID_ID_LENGHT, '0'));
        return sb.toString();
    }

    private ProcessDataVariable createProcessDataVariable(ValuePresentation presentation, DDI ddi) throws NumberFormatException {
        ProcessDataVariable processDataVariable = new ProcessDataVariable();
        processDataVariable.setProcessDataDDI(ddi.getDDEntityId()); // A
        processDataVariable.setProcessDataValue(0); // B
        processDataVariable.setValuePresentationIdRef(presentation);
        return processDataVariable;
    }

    private IsobusTreatmentZone createEmptyTreatmentZone() {
        IsobusTreatmentZone emptyTreatmentZone = new IsobusTreatmentZone();
        emptyTreatmentZone.setTreatmentZoneCode(EMPTY_TREATMENTZONE_CODE);
        emptyTreatmentZone.setTreatmentZoneDesignator(EMPTY_TREATMENTZONE_DESIGNATOR);
        emptyTreatmentZone.setTreatmentZoneColor(EMPTY_TREATMENTZONE_COLOR);
        return emptyTreatmentZone;
    }
    
    private IsobusTreatmentZone createDefaultTreatmentZone() {
        IsobusTreatmentZone defaultTreatmentZone = new IsobusTreatmentZone();
        defaultTreatmentZone.setTreatmentZoneCode(DEFAULT_TREATMENTZONE_CODE);
        defaultTreatmentZone.setTreatmentZoneDesignator(DEFAULT_TREATMENTZONE_DESIGNATOR);
        defaultTreatmentZone.setTreatmentZoneColor(DEFAULT_TREATMENTZONE_COLOR);
        return defaultTreatmentZone;
    }

    private DataLog createDataLog() throws NumberFormatException {
        DataLog dataLog = new DataLog();
        dataLog.setDDI(new byte[] { Byte.parseByte("-33"), Byte.parseByte("-1") }); // "DFFF"
        dataLog.setMethod((short) 31);
        dataLog.setTimeInterval(1000);
        return dataLog;
    }

    private IsobusTask createIsobusTask(String taskId) {
        IsobusTask isobusTask = new IsobusTask();
        isobusTask.setTaskId(taskId);
        isobusTask.setTaskStatus(TASK_STATUS);
        return isobusTask;
    }

    private ValuePresentation createValuePresentation() {
        ValuePresentation presentation = new ValuePresentation();
        presentation.setValuePresentationId(VALUE_PRESENTATION_ID);
        presentation.setOffset(VALUE_PRESENTATION_OFFSET);
        presentation.setScale(VALUE_PRESENTATION_SCALE);
        presentation.setNumberOfDecimals(VALUE_PRESENTATION_DECIMALS);
        presentation.setUnitDesignator(VALUE_PRESENTATION_DESIGNATOR);
        return presentation;
    }

    private IsobusTaskData createIsobusTaskData() {
        // create root object IsobusTaskData and fill in static config data
        IsobusTaskData isobusTaskData = new IsobusTaskData();
        isobusTaskData.setDataTransferOrigin(DATA_TRANSFER_ORIGIN);
        isobusTaskData.setManagementSoftwareManufacturer(MANAGEMENT_SOFTWARE);
        isobusTaskData.setTaskControllerVersion("1.0");
        return isobusTaskData;
    }
}
