/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.isobus.type;

import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import org.geotools.coverage.CoverageFactoryFinder;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.geometry.BoundingBox;

/**
 * Raster factory for Isobus grids
 * 
 * To create a raster it needs a IsobusTreatmentZoneGrid object and a IsobusTreatmentZoneGrid
 * It will assume that the treatment zones are in the right order
 * It assumes that element 0 == the right top corner of the grid
 * @author Merijn Zengers
 */
public class RasterFactory {


    /**
     * create a raster from grid data Calculates a raster from the information
     * in grid Then puts the values from the IsobusTreatmentZoneGrid in this raster 
     * Will take the first element of IsobusTreatmentZoneGrid as left top cell in the raster
     * Will then work from left to right and top to bottom through the raster
     * for each cell taking the next element from treatmentZones This requires
     * that grid.getMaximumColumn() * grid.getMaximumRow() ==
     * IsobusTreatmentZoneGrid.size() if this is not true will throw an
     * assertionException
     *
     * @param grid the grid to create a raster for
     * @param treatmentZones a list of treatmentZones for this grid
     * It will assume that the treatment zones are in the inverted order
     * It assumes that element 0 == the bottom right corner of the grid
     * and that element size()  == the top right corner
     * @return raster the raster created
     * @deprecated moving from TZN to value grid
     */
    @Deprecated
    public static GridCoverage2D createRaster(IsobusGrid grid, IsobusTreatmentZoneGrid treatmentZoneGrid) {
        assert (grid.getMaximumColumn() * grid.getMaximumRow() == treatmentZoneGrid.getSize());

        double minx = grid.getMinimumEastPosition().doubleValue();
        double maxx = grid.getMinimumEastPosition().doubleValue() + (grid.getCellEastSize() * grid.getMaximumColumn());
        double miny = grid.getMinimumNorthPosition().doubleValue() ;
        double maxy = grid.getMinimumNorthPosition().doubleValue() + (grid.getCellNorthSize() * grid.getMaximumRow());

        int rows = treatmentZoneGrid.getNumberOfRows();
        int columns = treatmentZoneGrid.getNumberOfColumn();
        int bands = 1;
        Point location = new Point(0, 0);
        WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, columns, rows, bands, location);

        for (int row = 0 ; row < rows; row++) {
            for (int column = 0; column < columns; column++) {
                Short treatMentZoneColor =  treatmentZoneGrid.getTreatmentZone(row, column).getTreatmentZoneColor() == null ? 0 : treatmentZoneGrid.getTreatmentZone(row, column).getTreatmentZoneColor();
                raster.setSample(column, row, 0, treatMentZoneColor);
            }
        }

        BoundingBox boundingBox = new ReferencedEnvelope(minx, maxx, miny, maxy, DefaultGeographicCRS.WGS84);


        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", raster, boundingBox);
    }
    
    /**
     * Creates a raster from a grid of raw process data values.
     * 
     * @param grid the grid to create a raster for (matches the grid tag in taskdata.xml)
     * @param valueGrid data grid of raw ProcessDataVariable values;
     * 
     * @return the raster created
     */
    public static GridCoverage2D createRaster(IsobusGrid grid, IsobusValueGrid valueGrid) {
        assert grid.getMaximumColumn() * grid.getMaximumRow() == valueGrid.getSize();

        double minx = grid.getMinimumEastPosition().doubleValue();
        double maxx = grid.getMinimumEastPosition().doubleValue() + (grid.getCellEastSize() * grid.getMaximumColumn());
        double miny = grid.getMinimumNorthPosition().doubleValue() ;
        double maxy = grid.getMinimumNorthPosition().doubleValue() + (grid.getCellNorthSize() * grid.getMaximumRow());

        int rows = valueGrid.getNumberOfRows();
        int columns = valueGrid.getNumberOfColumns();
        int bands = 1;
        Point location = new Point(0, 0);
        WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, columns, rows, bands, location);

        raster.setSamples(0, 0, columns, rows, 0, valueGrid.getValues());

        BoundingBox boundingBox = new ReferencedEnvelope(minx, maxx, miny, maxy, DefaultGeographicCRS.WGS84);

        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", raster, boundingBox);
    }
}
