/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.xml.parsers.ParserConfigurationException;
import nl.cloudfarming.client.isobus.model.*;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.FileEntry;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiDataObject.Entry;
import org.openide.loaders.MultiFileLoader;
import org.openide.util.NbBundle;
import org.xml.sax.SAXException;

/**
 * File loader for TASKDATA files TASKDATA files are in the format TASKDATA.XML
 * or this format in lower case
 *
 * Mandatory files : TASKDATA.XML or this format in lower case —the task data
 * file
 *
 * Optional files : TLG[0-9][0-9][0-9][0-9][0-9].BIN or this format in lower
 * case — contains time log values; TLG[0-9][0-9][0-9][0-9][0-9].XML or this
 * format in lower case — contains time log values;
 * GRD[0-9][0-9][0-9][0-9][0-9].BIN or this format in lower case — grid format;
 *
 * @author Merijn Zengers
 */
@NbBundle.Messages("TaskDataFileLoader timelog not found=could not find time log file {0} with exception{1}")
public class TaskDataFileLoader extends MultiFileLoader {

    public static final String PRIMARY_PATTERN = "[Tt][Aa][Ss][Kk][Dd][Aa][Tt][Aa]\\.[Xx][Mm][Ll]";
    public static final String GRID_PATTERN = "[Gg][Rr][Dd][0-9][0-9][0-9][0-9][0-9]\\.[Bb][Ii][Nn]";
    public static final String TIMELOG_PATTERN_BIN = "[Tt][Ll][Gg][0-9][0-9][0-9][0-9][0-9]\\.[Bb][Ii][Nn]";
    public static final String TIMELOG_PATTERN_XML = "[Tt][Ll][Gg][0-9][0-9][0-9][0-9][0-9]\\.[Xx][Mm][Ll]";
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type");

    public TaskDataFileLoader() {
        super(TaskDataObject.class.getName());
    }

    @Override
    protected String actionsContext() {
        StringBuilder builder = new StringBuilder("Loaders/");
        builder.append(TaskDataObject.MIME_TYPE);
        builder.append("/Actions/");
        return builder.toString();
    }

    /**
     * Finds the primary file by one of the two secondary files The secondary
     * files are GRD[0-9][0-9][0-9][0-9][0-9].BIN or toLower And
     * TLG[0-9][0-9][0-9][0-9][0-9].BIN or toLower
     *
     * @param fo the fileObject to find the primary file by
     * @return if fo is one of the secondary files returns the primary file of
     * format TASKDATA.XML or toLower
     */
    @Override
    protected FileObject findPrimaryFile(FileObject fo) {
        // never recognize folders.
        if (fo.isFolder()) {
            return null;
        }
        Pattern primaryPattern = Pattern.compile(PRIMARY_PATTERN);
        Pattern gridPattern = Pattern.compile(GRID_PATTERN);
        Pattern timeLogPatternBin = Pattern.compile(TIMELOG_PATTERN_BIN);
        Pattern timeLogPatternXml = Pattern.compile(TIMELOG_PATTERN_XML);
        if (primaryPattern.matcher(fo.getNameExt()).find()) {
            return fo;
        } else if (gridPattern.matcher(fo.getNameExt()).find() || timeLogPatternBin.matcher(fo.getNameExt()).find() || timeLogPatternXml.matcher(fo.getNameExt()).find()) {
            for (FileObject child : fo.getParent().getChildren()) {
                if (!child.equals(fo)) {
                    Pattern childPattern = Pattern.compile(PRIMARY_PATTERN);
                    if (childPattern.matcher(child.getNameExt()).find()) {
                        return child;
                    }
                }
            }
        }
        LOGGER.log(Level.FINEST, "No primary file found for FileObject {0}", fo);
        return null;
    }

    /**
     * Creates the MultidataObject for the file with format TASKDATA.XML or
     * toLower() Will add GRID and TimeLog files as secondary files if present
     * Only ads secondary files if they are referenced in the primary file
     *
     * @param primaryFile the Primary file to create the DataObject for
     * @return the created DataObject
     * @throws DataObjectExistsException if the DataObject already exists
     * @throws IOException
     */
    @Override
    protected MultiDataObject createMultiObject(FileObject primaryFile) throws DataObjectExistsException, IOException {
        List<FileObject> secondary = new ArrayList<>();
        IsobusTaskData isobusTaskData = new IsobusMarshaller().unMarshall(primaryFile);
        // Retrieve timeLogs into isobusTaskData
        //See TODO on method
        //        if (isobusTaskData != null) {
        //          addTimelogs(isobusTaskData, primaryFile);
        //        }
        for (FileObject child : primaryFile.getParent().getChildren()) {
            if (!child.equals(primaryFile)) {
                //Check every task in the taskadata for grid files
                for (IsobusTask task : isobusTaskData.getTSK()) {
                    if (task.getGRD() != null && task.getGRD().getFilename().equalsIgnoreCase(child.getName())) {
                        secondary.add(child);
                    }
                    for (TimeLog timeLog : task.getTLG()) {
                        if (timeLog.getFilename().equalsIgnoreCase(child.getName())) {
                            secondary.add(child);
                        }
                    }
                }

            }
        }
        TaskDataObject dataObject = new TaskDataObject(primaryFile, this, isobusTaskData, secondary);
        return dataObject;
    }

    /**
     * Creates a primary entry for the primary files
     *
     * @param obj the dataObject to create a primary entry for
     * @param primaryFile the primary file to add as primary entry
     * @return the created primary entry
     */
    @Override
    protected Entry createPrimaryEntry(MultiDataObject obj, FileObject primaryFile) {
        Entry entry = new FileEntry(obj, primaryFile);
        LOGGER.log(Level.FINEST, "Creating primary entry for {0}", primaryFile.getName());
        return entry;
    }

    /**
     * Creates a secondary entry for the provided MultiDataObject
     *
     * @param obj the MultiDataObject to create the secondary entry for
     * @param secondaryFile the secondary file to create the entry from
     * @return The created entry
     */
    @Override
    protected Entry createSecondaryEntry(MultiDataObject obj, FileObject secondaryFile) {
        Pattern gridPattern = Pattern.compile(GRID_PATTERN);
        Pattern timeLogPatternBin = Pattern.compile(TIMELOG_PATTERN_BIN);
        Pattern timeLogPatternXml = Pattern.compile(TIMELOG_PATTERN_XML);
        assert (gridPattern.matcher(secondaryFile.getNameExt()).find() || timeLogPatternBin.matcher(secondaryFile.getNameExt()).find() || timeLogPatternXml.matcher(secondaryFile.getNameExt()).find());

        FileEntry entry = new FileEntry(obj, secondaryFile);
        LOGGER.log(Level.FINEST, "Creating secondary entry for {0}", secondaryFile.getName());
        return entry;
    }

    /**
     * Will read Time objects from binary file + xml file. Will get timeLog
     * property from the tasks When TimeLog.getFileName is not null a new
     * TimeLogReader will be created The FileName and the Directory which must
     * contain a XML file header and a binary file will be parsed to the
     * TimeLogReader. The FileName of the XML file must be TimeLog.fileName with
     * a extension of .xml or .XML The FileName of the binary file must be
     * TimeLog.fileName with a extension of .bin or .BIN
     *
     * @param isobusTask the Isobus object containing the tasks and in turn the
     * TimeLog objects
     * @param xmlFile The XML-file, needed to retrieve timelog from the same
     * folder.
     * @return a List of times read from the XML and binary files combined.
     * @throws ParserConfigurationException
     * @throws SAXException
     * @throws IOException
     */
    //TODO: Made wrong assumption that times need to be added to the task
    //Time element of task is meant to store the start and end time of task in
    //Probably need a Timelog fieloader and visual element to visualize 
    private void addTimelogs(IsobusTaskData isobusTaskData, FileObject xmlFile) {
        String timeLogFileName = "";
        FileObject parentDir = xmlFile.getParent();
        for (IsobusTask isobusTask : isobusTaskData.getTSK()) {
            List<IsobusTime> times = new ArrayList<>();
            try {
                for (TimeLog timeLog : isobusTask.getTLG()) {
                    timeLogFileName = timeLog.getFilename();
                    if (timeLogFileName != null && !timeLogFileName.equals("")) {
                        TimeLogReader timeLogReader = new TimeLogReader(timeLogFileName, parentDir, isobusTaskData);
                        times.addAll(timeLogReader.getTimes());
                    }
                }
                isobusTask.getTIM().addAll(times);
            } catch (FileNotFoundException ex) {
                LOGGER.log(Level.SEVERE, Bundle.TaskDataFileLoader_timelog_not_found(timeLogFileName, ex));
            } catch (IOException | SAXException | ParserConfigurationException ex) {
                throw new IsobusUnmarshallingException(NbBundle.getMessage(IsobusMarshaller.class, "isobus.error.isobus_unmarshaller.exception", xmlFile.getName()) + ex);
            }
        }
    }
}
