/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.isobus.type;

import nl.cloudfarming.client.isobus.tree.TaskDataRootNode;
import java.io.IOException;
import java.util.List;
import java.util.regex.Pattern;
import nl.cloudfarming.client.isobus.model.*;
import org.netbeans.core.spi.multiview.MultiViewElement;
import org.netbeans.core.spi.multiview.text.MultiViewEditorElement;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.Node;
import org.openide.util.HelpCtx;
import org.openide.util.Lookup;
import org.openide.util.NbBundle.Messages;
import org.openide.windows.TopComponent;

/**
 * DataObject representing isobus taskdata.
 *
 * <p>The TaskDataObject is a MultiDataObject, the TASKDATA.XML contain
 * references to binary files for grids and timelogs. Those linked binary files
 * are contained as secondary files in this dataobject. In order to be able to
 * link the secondary files, the taskdata,xml file needs to be parsed, which can
 * be slow.</p>
 *
 * @author Timon Veenstra <monezz@gmail.com>
 */
public class TaskDataObject extends MultiDataObject {

    public static final String MIME_TYPE = "application/x-taskdata";
    private IsobusTaskData isobusTaskData;

    public TaskDataObject(FileObject pf, MultiFileLoader loader, IsobusTaskData isobusTaskData, List<FileObject> secondary) throws DataObjectExistsException, IOException {
        super(pf, loader);


        registerEditor(MIME_TYPE, true);
        this.isobusTaskData = isobusTaskData;

        setSecondaryFiles(secondary);
    }

    /**
     * Method to set the secondary files
     *
     * @param secondary the secondary files
     */
    private void setSecondaryFiles(List<FileObject> secondary) {
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }
    }

    /**
     * Get the IsobusTaskData from this DataObject
     *
     * @return IsobusTaskData object for this DataObject
     */
    public IsobusTaskData getIsobusTaskData() {
        return this.isobusTaskData;
    }

    /**
     * Gets the isobus grid object from the taskdata matching the supplied grid
     * binary file Will throw AssertionExceptions if the supplied file does not
     * match the grid filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN or if
     * the supplied file is not a secondary file of this TaskDataObject
     *
     * @param gridFile the grid binary file to search the isobus grid for
     * @return The IsobusTreatmentZoneGrid object if found else null
     */
    public IsobusGrid getGrid(FileObject gridFile) {
        assert findSecondaryEntry(gridFile) != null : "The supplied file is not secondary file of this TaskDataDataObject";
        Pattern gridPattern = Pattern.compile(TaskDataFileLoader.GRID_PATTERN);
        assert gridPattern.matcher(gridFile.getNameExt()).find() : "The file name of the Suplied file does not match the grid filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN";
        IsobusGrid grid = null;
        for (IsobusTask task : isobusTaskData.getTSK()) {
            if (task.getGRD() != null && task.getGRD().getFilename() != null && task.getGRD().getFilename().toLowerCase().equals(gridFile.getName().toLowerCase())) {
                grid = task.getGRD();
            }
        }
        return grid;
    }

    /**
     * Gets the isobus treatment zones referenced in the supplied grid file Will
     * throw AssertionExceptions if the supplied file does not match the grid
     * filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN or if the supplied
     * file is not a secondary file of this TaskDataObject
     *
     * @param gridFile the grid file to search the treatment zones for
     * @return a list of found treatment zones or null if nothing is found
     */
    public List<IsobusTreatmentZone> getTreatmentzones(FileObject gridFile) {
        assert findSecondaryEntry(gridFile) != null : "The supplied file is not secondary file of this TaskDataDataObject";
        Pattern gridPattern = Pattern.compile(TaskDataFileLoader.GRID_PATTERN);
        assert gridPattern.matcher(gridFile.getNameExt()).find() : "The file name of the Suplied file does not match the grid filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN";
        List<IsobusTreatmentZone> treatmentZones = null;
        for (IsobusTask task : isobusTaskData.getTSK()) {
            if (task.getGRD() != null && task.getGRD().getFilename() != null && task.getGRD().getFilename().toLowerCase().equals(gridFile.getName().toLowerCase())) {
                treatmentZones = task.getTZN();
            }
        }
        return treatmentZones;
    }

    @Override
    protected Node createNodeDelegate() {
        //FIXME quick workaround for AGROSENSE-582, investigate and apply proper fix
        if (isobusTaskData != null) {
            return new TaskDataRootNode(isobusTaskData, super.createNodeDelegate());
        } else {
            return super.createNodeDelegate();
        }
    }

    @Override
    public HelpCtx getHelpCtx() {
        return new HelpCtx("nl.cloudfarming.client.isobus.project.explorer");
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    @MultiViewElement.Registration(displayName = "#LBL_Isobus_EDITOR",
    iconBase = "nl/cloudfarming/client/isobus/type/iso.png",
    mimeType = "application/x-taskdata",
    persistenceType = TopComponent.PERSISTENCE_ONLY_OPENED,
    preferredID = "Isobus",
    position = 1000)
    @Messages("LBL_Isobus_EDITOR=Source")
    public static MultiViewEditorElement createEditor(Lookup lkp) {
        return new MultiViewEditorElement(lkp);
    }
}
