/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.Closeable;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.util.*;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import nl.cloudfarming.client.isobus.model.BinaryReader;
import nl.cloudfarming.client.isobus.model.DataLogValue;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.IsobusDevice;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.Position;
import nl.cloudfarming.client.isobus.model.PositionStatusEnum;
import nl.cloudfarming.client.isobus.model.TimeTypeEnum;
import nl.cloudfarming.client.util.BinaryUtil;
import org.openide.filesystems.FileObject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Class that will construct Time Object from a binary file and the XML-file
 * Constructor expects the fileName without extension and a file object which
 * represents the map containing the files
 *
 * @author Merijn Zengers & Gerben Feenstra
 */
public class TimeLogReader implements Closeable {

    private XmlReader xmlReader;
    private BinaryReader binaryReader;
    private IsobusTaskData isobusTaskData;

    public TimeLogReader(String timeLogFileName, FileObject parentDir, IsobusTaskData isobusTaskData) throws ParserConfigurationException, SAXException, FileNotFoundException, IOException {
        this.isobusTaskData = isobusTaskData;

        FileObject timeLogXml = getTimeLogXml(timeLogFileName, parentDir);
        FileObject timeLogBin = getTimeLogBin(timeLogFileName, parentDir);

        binaryReader = new BinaryReader(timeLogBin.getInputStream());
        xmlReader = new XmlReader(timeLogXml.getInputStream());
    }

    /**
     * Get the time elements from the binary file
     *
     * @return a list of times constructed from the XML and the binary file
     * combined
     * @throws IOException when the binary or the XML file cannot be read
     */
    public List<IsobusTime> getTimes() throws IOException {
        ArrayList<IsobusTime> times = new ArrayList<>();
        if (binaryReader != null && xmlReader != null) {
            while (binaryReader.available()) {
                IsobusTime time = new IsobusTime();
                Map<TimeFieldEnum, String> timeMap = xmlReader.getTimeElement();

                // Get 01-01-1980
                Calendar cal = getIsobusEpoch();

                for (TimeFieldEnum timeFieldEnum : timeMap.keySet()) {

                    // Empty attributes will be read from binary
                    switch (timeFieldEnum) {
                        case TIME_START_TIME:
                            if ("".equals(timeMap.get(timeFieldEnum))) {
                                cal.add(Calendar.MILLISECOND, Long.valueOf(binaryReader.read(timeFieldEnum)).intValue());
                            }
                            // else -> Can starttime be specified in XML?
                            break;
                        case TIME_START_DATE:
                            if ("".equals(timeMap.get(timeFieldEnum))) {
                                cal.add(Calendar.DAY_OF_YEAR, Long.valueOf(binaryReader.read(timeFieldEnum)).intValue());
                            }
                            // else -> Can startdate be specified in XML?
                            break;
                        case TIME_TYPE:
                            if ("".equals(timeMap.get(timeFieldEnum))) {
                                int timeType = Long.valueOf(binaryReader.read(timeFieldEnum)).intValue();
                                // TimeTypeEnum starts at 1 instead of 0, so do a -1
                                time.setTimeType(TimeTypeEnum.values()[timeType-1]);
                                //time.setTimeType(getTimeType(Long.valueOf(binaryReader.read(timeFieldEnum)).intValue()));
                            } else {
                                int timeType = Integer.parseInt(timeMap.get(timeFieldEnum));  
                                // TimeTypeEnum starts at 1 instead of 0, so do a -1
                                time.setTimeType(TimeTypeEnum.values()[timeType-1]);
                            }
                    }
                }
                time.setTimeStart(cal.getTime());
                times.add(time);

                //Position
                time.setPTN(getPosition(xmlReader.getPositionElement()));
                
                //DLV
                time.getDLV().addAll(getDataLogValues(xmlReader.getDataLogValues()));
            }
        }
        return times;
    }

    /**
     * Gets the position from binary file based on supplied Map of
     * PositionFieldEnum and String
     *
     * @param positionMap the position map that contains the PositionFieldEnum
     * and a string containing the value for this PositionFieldEnum
     * @return the position read from the binary file
     * @throws IOException when the binary file can not be read
     */
    private Position getPosition(Map<PositionFieldEnum, String> positionMap) throws IOException {
        Position position = new Position();
        for (PositionFieldEnum positionFieldEnum : positionMap.keySet()) {
            switch (positionFieldEnum) {
                case POSITION_NORTH:
                    position.setPositionNorth(BigDecimal.valueOf(binaryReader.read(positionFieldEnum) / 10000000d));
                    break;
                case POSITION_EAST:
                    position.setPositionEast(BigDecimal.valueOf(binaryReader.read(positionFieldEnum) / 10000000d));
                    break;
                case POSITION_UP:
                    position.setPositionUp(Long.valueOf(binaryReader.read(positionFieldEnum)).intValue());
                    break;
                case POSITION_STATUS:
                    position.setPositionStatus(getPositionStatus(Double.valueOf(binaryReader.read(positionFieldEnum)).intValue()));
                    break;
                case POSITION_HDOP:
                    position.setHDOP(new BigDecimal(binaryReader.read(positionFieldEnum)));
                    break;
                case POSITION_PDOP:
                    position.setPDOP(new BigDecimal(binaryReader.read(positionFieldEnum)));
                    break;
                case POSITION_SAT_NR:
                    position.setNumberOfSatellites(Long.valueOf(binaryReader.read(positionFieldEnum)).shortValue());
                    break;
                case POSITION_GPS_UTC_TIME:
                    position.setGpsUtcTime(binaryReader.read(positionFieldEnum));
                    break;
                case POSITION_GPS_UTC_DATE:
                    position.setGpsUtcDate(Long.valueOf(binaryReader.read(positionFieldEnum)).intValue());
                    break;
            }
        }
        return position;
    }

    /**
     * Gets the DatalogValues from the binary file for the provided List of maps
     * containing DataLogfieldEnums and strings
     *
     * @param dataLogValueMaps the list of maps to get the DataLogValues for
     * @return a list of DataLogValues read from the binary file
     * @throws IOException when binary file can not be read
     */
    private List<DataLogValue> getDataLogValues(List<Map<DatalogValueFieldEnum, String>> dataLogValueMaps) throws IOException {
        //DataLogValue
        List<DataLogValue> dataLogValues = new ArrayList<>();
        int dataLogValueIndexXml = 0;
        int dataLogValueIndexBin = 0;
        int dataLogValueCount = 0;
        for (Map<DatalogValueFieldEnum, String> dataLogValueMap : dataLogValueMaps) {
            DataLogValue dataLogValue = new DataLogValue();
            for (DatalogValueFieldEnum datalogFieldEnum : dataLogValueMap.keySet()) {
                switch (datalogFieldEnum) {
                    case PROCESS_DATA_DDI:
                        String hexAsString = dataLogValueMap.get(datalogFieldEnum);
                        int hexAsInt = Integer.parseInt(hexAsString, 16);

                        byte[] bytes = BinaryUtil.intToByteArray(hexAsInt);
                        dataLogValue.setProcessDataDDI(bytes);
                        break;
                    case DEVICE_ELEMENT_ID_REF:
                        dataLogValue.setDeviceElementIdRef(findDeviceElementByID(dataLogValueMap.get(datalogFieldEnum)));
                        break;
                }
            }
            if (dataLogValueCount == 0) {
                dataLogValueCount = Long.valueOf(binaryReader.read(DatalogValueFieldEnum.DATALOGVALUE_COUNT)).intValue();
            }
            if (dataLogValueIndexXml > dataLogValueIndexBin || dataLogValueIndexBin == 0) {
                dataLogValueIndexBin = Long.valueOf(binaryReader.read(DatalogValueFieldEnum.DATALOGVALUE_NR)).intValue();
            }
            if (dataLogValueIndexXml == dataLogValueIndexBin) {
                dataLogValue.setProcessDataValue(Long.valueOf(binaryReader.read(DatalogValueFieldEnum.DATALOGVALUE_PROCESSDATA_VALUE)).intValue());
            }
            dataLogValueIndexXml++;
            dataLogValues.add(dataLogValue);
        }
        return dataLogValues;
    }

    /**
     * Helper method to find device by ID
     * @param id Id of the device
     * @return Device with the specified ID
     */
    protected DeviceElement findDeviceElementByID(String id) {
        for (IsobusDevice device : isobusTaskData.getDVC()) {
            for (DeviceElement deviceElement : device.getDET()) {
                if (deviceElement.getId().equals(id)) {
                    return deviceElement;
                }
            }
        }
        return null;
    }

    @Override
    public void close() throws IOException {
        binaryReader.close();
        binaryReader = null;
        // TODO: Make XMLReader closeable
    }

    /**
     * Class to read the header of a binary timeLog file (header is a
     * name-equivalent XML-file)
     *
     * @author Gerben Feenstra & Merijn Zengers
     */
    private static class XmlReader {

        private Document xmlDoc = null;
        private Map<TimeFieldEnum, String> timeMap;
        private Map<PositionFieldEnum, String> positionMap;
        private LinkedList<Map<DatalogValueFieldEnum, String>> dataLogValues;
        private Element timeLogRootElement;

        public XmlReader(InputStream xmlInputStream) throws ParserConfigurationException, SAXException, FileNotFoundException {

            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            try {
                xmlDoc = docBuilder.parse(xmlInputStream);
            } catch (IOException ex) {
                throw new FileNotFoundException(ex.getMessage());
            }

            // Get root element (TIM)
            timeLogRootElement = xmlDoc.getDocumentElement();
        }

        /**
         * Gets the Time element from XML Will construct a LinkedHashMap
         * containing a TimeFieldEnum as key and the value from XML as value if
         * the map is already constructed returns the already constructed map
         *
         * @return the constructed map
         */
        Map<TimeFieldEnum, String> getTimeElement() {
            if (timeMap == null) {
                // Keep map ordered by natural ordering
                timeMap = new LinkedHashMap<>();

                for (TimeFieldEnum timeFieldEnum : TimeFieldEnum.values()) {
                    // Only attributes that ARE PRESENT are stored in the binary file belonging to the XML
                    // So we explicitly check for the existence of the attribute
                    if (timeLogRootElement.hasAttribute(timeFieldEnum.getAttributeName())) {
                        String value = timeLogRootElement.getAttribute(timeFieldEnum.getAttributeName());
                        // Empty value will be parsed from the binary file using the binary reader.
                        timeMap.put(timeFieldEnum, value);
                    }
                }
            }
            return timeMap;
        }

        /**
         * Gets the Position element from XML Will construct a Linked HashMap
         * containing a PositionFieldEnum as key and the value from the XML as
         * value if the Map is already constructed will return the already
         * constructed map
         *
         * @return the constructed LinkedHashMap
         */
        Map<PositionFieldEnum, String> getPositionElement() {
            if (positionMap == null) {

                positionMap = new LinkedHashMap<>();
                // Get positions list
                NodeList position = timeLogRootElement.getElementsByTagName("PTN");
                Node positionNode;
                Element positionElement = null;
                // Retrieve the Position Element
                if (position.getLength() > 0) {
                    positionNode = position.item(0);
                    if (positionNode.getNodeType() == Node.ELEMENT_NODE) {
                        positionElement = (Element) positionNode;
                    }
                }

                for (PositionFieldEnum fieldEnum : PositionFieldEnum.values()) {
                    if (positionElement.hasAttribute(fieldEnum.getAttributeName())) {
                        String value = positionElement.getAttribute(fieldEnum.getAttributeName());
                        positionMap.put(fieldEnum, value);
                    }
                }
            }
            return positionMap;
        }

        /**
         * Gets the data log values from the XML Will construct a Linked list
         * containing maps with DatalogValueFieldEnum as key and the value from
         * the XML as value If the linked list is already constructed the list
         * is returned instantly
         *
         * @return The constructed linked list
         */
        List<Map<DatalogValueFieldEnum, String>> getDataLogValues() {
            if (dataLogValues == null) {
                dataLogValues = new LinkedList<>();
                // Get datalogvalues list
                NodeList datalogNodes = timeLogRootElement.getElementsByTagName("DLV");
                Node datalogValueNode;
                Element datalogValue = null;

                for (int i = 0; i < datalogNodes.getLength(); i++) {
                    LinkedHashMap dataLogValueMap = new LinkedHashMap<>();
                    datalogValueNode = datalogNodes.item(i);
                    if (datalogValueNode.getNodeType() == Node.ELEMENT_NODE) {
                        datalogValue = (Element) datalogValueNode;
                    }

                    for (DatalogValueFieldEnum dataLogEnum : DatalogValueFieldEnum.values()) {
                        if (datalogValue.hasAttribute(dataLogEnum.getAttributeName())) {
                            String value = datalogValue.getAttribute(dataLogEnum.getAttributeName());
                            dataLogValueMap.put(dataLogEnum, value);
                        }
                    }
                    dataLogValues.add(dataLogValueMap);
                }
            }
            return dataLogValues;
        }
    }

    /**
     * TODO: Refactor to something nicer. Method to get the binary file from the
     * specified directory Will only accept files with the fileName provided to
     * the constructor with a .bin or .BIN extension
     *
     * @param dir the directory where the file should reside in
     * @return the File if found else null
     */
    private FileObject getTimeLogBin(final String fileName, final FileObject dirObj) throws IOException {
        FileObject result;
        String fileCap = fileName + ".BIN";
        result = dirObj.getFileObject(fileCap);

        if (result == null) {
            String fileNoCap = fileName + ".bin";
            result = dirObj.getFileObject(fileNoCap);
        }

        if (result == null) {
            throw new IOException("Binary Timelog " + result + " could not be found");
        }
        return result;
    }

    /**
     * TODO: Refactor to something nicer. Gets the XML file from the specified
     * directory Will only accept files with the fileName provided to the
     * constructor with a .xml or .XML extension
     *
     * @param dir the directory to find the file in
     * @return null if not found else the the File object
     */
    private FileObject getTimeLogXml(final String fileName, final FileObject dirObj) throws IOException {
        FileObject result;
        String fileCap = fileName + ".XML";
        result = dirObj.getFileObject(fileCap);

        if (result == null) {
            String fileNoCap = fileName + ".xml";
            result = dirObj.getFileObject(fileNoCap);
        }

        if (result == null) {
            throw new IOException("XML Timelog " + result + " could not be found");
        }
        return result;
    }

    /**
     * TODO: Remove!
     *
     * @deprecated Replace by PositionStatusEnum.values()[value] ??
     */
    @Deprecated
    private PositionStatusEnum getPositionStatus(int value) {
        switch (value) {
            case 0:
                return PositionStatusEnum.NO_GPS_FIX;
            case 1:
                return PositionStatusEnum.GNSS_FIX;
            case 2:
                return PositionStatusEnum.DGNSS_FIX;
            case 3:
                return PositionStatusEnum.PRECISE_GNSS;
            case 4:
                return PositionStatusEnum.RTK_FIXED_INTEGER;
            case 5:
                return PositionStatusEnum.RTK_FLOAT;
            case 6:
                return PositionStatusEnum.EST_DR_MODE;
            case 7:
                return PositionStatusEnum.MANUAL_INPUT;
            case 8:
                return PositionStatusEnum.SIMULATE_MODE;
            case 9:
            case 10:
            case 11:
            case 12:
            case 13:
                return PositionStatusEnum.RESERVED_1;
            case 14:
                return PositionStatusEnum.ERROR;
            case 15:
                return PositionStatusEnum.POSITION_STATUS_NOT_AVAILABLE;
        }
        return null;
    }

    /**
     * Creates a new Calendar object and set it on the first of januari 1980.
     *
     * @return A calendar object set to 01-01-1980
     */
    private Calendar getIsobusEpoch() {
        //make Calendar and set it on the first of januari 1980
        Calendar cal = Calendar.getInstance();
        cal.set(Calendar.YEAR, 1980);
        cal.set(Calendar.MONTH, 0);
        cal.set(Calendar.DAY_OF_MONTH, 1);
        cal.set(Calendar.HOUR_OF_DAY, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        return cal;
    }
}
